
""" A key-value store implementation based on sqlite3. """

import contextlib
import math
import sqlite3
import string
import threading
import time
import typing as t

from ._api import KeyValueStore


def _fetch_all(cursor: sqlite3.Cursor) -> t.Iterable[tuple]:
  while True:
    rows = cursor.fetchmany()
    print(rows)
    if not rows:
      break
    yield from rows


class SqliteDatastore:
  """
  Provider for key-value datastores backed by an SQLite3 database.

  The #SqliteDatastore is thread-safe, but may be slow to access concurrently due to locking requirements.
  """

  NAMESPACE_CHARS = frozenset(string.ascii_letters + string.digits + '._-')

  def __init__(self, filename: str) -> None:
    self._lock = threading.Lock()
    self._conn = sqlite3.connect(filename, check_same_thread=False)
    self._created_namespaces: t.Set[str] = set()

  @staticmethod
  def _get_time(add: int) -> int:
    return int(math.ceil(time.time() + add))

  @staticmethod
  def _validate_namespace(namespace: str) -> None:
    if set(namespace) - SqliteDatastore.NAMESPACE_CHARS:
      raise ValueError(f'invalid namespace name: {namespace!r}')

  @staticmethod
  def _get_namespaces(cursor: sqlite3.Cursor) -> t.Iterable[str]:
    cursor.execute('''SELECT name FROM sqlite_master
      WHERE type = 'table' AND name NOT like 'sqlite_%';''')
    yield from (x[0] for x in _fetch_all(cursor))

  @contextlib.contextmanager
  def _locked_cursor(self) -> t.Iterator[sqlite3.Cursor]:
    with self._lock, contextlib.closing(self._conn.cursor()) as cursor:
      yield cursor

  def get_namespaces(self) -> t.Iterator[str]:
    """
    Returns an iterator that returns the name of all namespaces known to the Sqlite store. Note
    that new namespaces are created on-deman using #store().
    """

    with self._locked_cursor() as cursor:
      yield from self._get_namespaces(cursor)

  def get_keys(self, namespace: str, prefix: str) -> t.Iterator[tuple[str, int | None]]:
    """
    Returns an iterator that returns all keys in the specified *namespace* and their expiration
    timestamp. This excludes any keys that are already expired but not yet expunged.
    """

    with self._locked_cursor() as cursor:
      try:
        cursor.execute(f'''
          SELECT key, exp FROM "{namespace}"
          WHERE key LIKE ? AND (exp IS NULL OR exp < ?)
          ''',
          (prefix.replace('%', '%%') + '%', self._get_time(0),))
      except sqlite3.OperationalError as exc:
        if 'no such table' in str(exc):
          raise ValueError(f'namespace {namespace!r} does not exist')
        raise
      yield from t.cast(t.Iterator[tuple[str, int | None]], _fetch_all(cursor))

  def _ensure_namespace(self, cursor: sqlite3.Cursor, namespace: str) -> None:
    self._validate_namespace(namespace)
    if namespace not in self._created_namespaces:
      cursor.execute(f'''
        CREATE TABLE IF NOT EXISTS "{namespace}"
        (key TEXT PRIMARY KEY, value BLOB, exp INTEGER)''')
      self._created_namespaces.add(namespace)

  def get(self, namespace: str, key: str) -> bytes:
    self._validate_namespace(namespace)
    with self._locked_cursor() as cursor:
      try:
        cursor.execute(f'''
          SELECT value FROM "{namespace}"
            WHERE key = ? AND (? < exp OR exp IS NULL)''',
          (key, self._get_time(0)),
        )
      except sqlite3.OperationalError as exc:
        if 'no such table' in str(exc):
          raise ValueError(f'namespace {namespace!r} does not exist')
        raise
      result = cursor.fetchone()
      if result is None:
          raise KeyError(f'{namespace} :: {key}')
      if not isinstance(result[0], bytes):
        raise RuntimeError(f'expected data to be bytes, got {type(result[0]).__name__}')
      return result[0]

  def set(self, namespace: str, key: str, value: bytes, expires_in: int | None = None) -> None:
    self._validate_namespace(namespace)
    with self._locked_cursor() as cursor:

      # Create the table for the namespace if it does not exist.
      self._ensure_namespace(cursor, namespace)

      # Insert the value into the database.
      exp = self._get_time(expires_in) if expires_in is not None else None
      cursor.execute(f'''
        INSERT OR REPLACE INTO "{namespace}"
        VALUES (?, ?, ?)''',
        (key, value, exp),
      )

      self._conn.commit()

  def delete(self, namespace: str, key: str) -> None:
    self._validate_namespace(namespace)
    with self._locked_cursor() as cursor:

      # Create the table for the namespace if it does not exist.
      self._ensure_namespace(cursor, namespace)

      # Insert the value into the database.
      cursor.execute(f'DELETE FROM "{namespace}" WHERE key = ?', (key,))

      self._conn.commit()

  def get_namespace(self, namespace: str) -> KeyValueStore:
    with self._locked_cursor() as cursor:
      self._ensure_namespace(cursor, namespace)
    return SqliteNamespace(self, namespace)

  def expunge(self, namespace: t.Optional[str] = None) -> None:
    with self._locked_cursor() as cursor:
      for namespace in [namespace] if namespace else list(self._get_namespaces(cursor)):
        cursor.execute(f'''
          DELETE FROM "{namespace}" WHERE exp < ?''',
          (self._get_time(0),),
        )
      self._conn.commit()


class SqliteNamespace(KeyValueStore):

  def __init__(self, store: SqliteDatastore, namespace: str) -> None:
    self._store = store
    self._namespace = namespace

  def get(self, key: str) -> bytes:
    return self._store.get(self._namespace, key)

  def set(self, key: str, value: bytes, expires_in: int | None = None) -> None:
    self._store.set(self._namespace, key, value, expires_in)

  def delete(self, key: str) -> None:
    self._store.delete(self._namespace, key)

  def keys(self, prefix: str = '') -> t.Iterable[str]:
    for key, _exp in self._store.get_keys(self._namespace, prefix):
      yield key

  def count(self, prefix: str = '') -> int:
    raise NotImplementedError('SqliteNamespace.count() is not implemented')




import typing as t

from nr.util.generic import T, K, V

T_ChainDict = t.TypeVar('T_ChainDict', bound='ChainDict')
_can_iteritems = lambda x: hasattr(x, 'items')


class ChainDict(t.MutableMapping[K, V]):
  """
  A dictionary that wraps a list of dictionaries. Except for the first dictionary passed into the #ChainDict
  constructor, none will be mutated through any actions on this dictionary.
  """

  def __init__(self, main: t.MutableMapping[K, V], *others: t.Mapping[K, V]) -> None:
    self._main = main
    self._dicts: t.List[t.Mapping[K, V]] = [t.cast(t.Mapping[K, V], main)] + list(others)
    self._deleted: t.Set[K] = set()
    self._in_repr = False

  def __contains__(self, key: t.Any) -> bool:
    if key not in self._deleted:
      for d in self._dicts:
        if key in d:
          return True
    return False

  def __getitem__(self, key: K) -> V:
    if key not in self._deleted:
      for d in self._dicts:
        try: return d[key]
        except KeyError: pass
    raise KeyError(key)

  def __setitem__(self, key: K, value: V) -> None:
    self._main[key] = value
    self._deleted.discard(key)

  def __delitem__(self, key: K) -> None:
    if key not in self:
      raise KeyError(key)
    try: self._main.pop(key)
    except KeyError: pass
    self._deleted.add(key)

  def __iter__(self) -> t.Iterator[K]:
    return self.keys()

  def __len__(self) -> int:
    return sum(1 for x in self.keys())

  def __repr__(self) -> str:
    if self._in_repr:
      return 'ChainDict(...)'
    else:
      self._in_repr = True
      try:
        return 'ChainDict({})'.format(dict(self.items()))
      finally:
        self._in_repr = False

  def __eq__(self, other: t.Any) -> bool:
    return dict(self.items()) == other

  def __ne__(self, other: t.Any) -> bool:
    return not (self == other)

  @t.overload
  def get(self, key: K) -> t.Optional[V]: ...

  @t.overload
  def get(self, key: K, default: T) -> t.Union[V, T]: ...

  def get(self, key, default=None):
    try:
      return self[key]
    except KeyError:
      return default

  @t.overload  # type: ignore  # TODO (NiklasRosenstein)
  def pop(self, key: K) -> V: ...

  @t.overload
  def pop(self, key: K, default: t.Union[V, T]) -> t.Union[V, T]: ...

  def pop(self, key, default=NotImplemented):
    try:
      value = self[key]
    except KeyError:
      if default is NotImplemented:
        raise KeyError(key)
      return default
    else:
      del self[key]
    return value

  def popitem(self) -> t.Tuple[K, V]:
    if self._main:
      key, value = self._main.popitem()
      self._deleted.add(key)
      return key, value
    for d in self._dicts:
      for key in d.keys():
        if key not in self._deleted:
          self._deleted.add(key)
          return key, d[key]
    raise KeyError('popitem(): dictionary is empty')

  def clear(self) -> None:
    self._main.clear()
    self._deleted.update(self.keys())

  def copy(self: T_ChainDict) -> T_ChainDict:
    return type(self)(self._main, *self._dicts[1:])

  def setdefault(self, key: K, value: V) -> V:  # type: ignore  # TODO (NiklasRosenstein)
    try:
      return self[key]
    except KeyError:
      self[key] = value
      return value

  def update(self, E: t.Mapping[K, V], *F: t.Mapping[K, V]) -> None:  # type: ignore  # TODO (NiklasRosenstein)
    if _can_iteritems(E):
      for k, v in E.items():
        self[k] = v
    else:
      for k in E.keys():
        self[k] = E[k]
    for Fv in F:
      for k, v in Fv.items():
        self[k] = v

  def keys(self):
    seen = set()
    for d in self._dicts:
      for key in d.keys():
        if key not in seen and key not in self._deleted:
          yield key
          seen.add(key)

  def values(self):
    seen = set()
    for d in self._dicts:
      for key, value in d.items():
        if key not in seen and key not in self._deleted:
          yield value
          seen.add(key)

  def items(self):
    seen = set()
    for d in self._dicts:
      for key, value in d.items():
        if key not in seen and key not in self._deleted:
          yield key, value
          seen.add(key)




import typing as t


class ASGIApp(t.Protocol):

  def __call__(self, scope, receive, send):
    raise NotImplementedError


class WSGIApp(t.Protocol):

  def __call__(self, environ, start_response):
    raise NotImplementedError




import logging
import typing as t
from pathlib import Path

import databind.json  # type: ignore[import]
import logging
import yaml  # type: ignore[import]
from databind.core.annotations import collect_unknowns  # type: ignore[import]

from nr.util.generic import T

logger = logging.getLogger(__name__)


class ConfigLoader:

  def __init__(
    self,
    filename: str,
    encoding: str = 'utf8',
    allow_unknown_keys: bool = False,
  ) -> None:
    self.filename = Path(filename)
    self.encoding = encoding
    self.allow_unknown_keys = allow_unknown_keys
    self.last_load_mtime: t.Optional[float] = None

  def load_config(self, model: type[T]) -> T:
    data = yaml.safe_load(self.filename.read_text(encoding=self.encoding))
    unknowns = collect_unknowns()
    annotations = [unknowns] if self.allow_unknown_keys else []
    config = databind.json.load(data, model, filename=str(self.filename), annotations=annotations)
    if unknowns.entries:
      logger.warning('Found unknown configuration keys in "%s": %s', self.filename, unknowns.entries)
    self.last_load_mtime = self.filename.stat().st_mtime
    return config

  def changed(self) -> bool:
    if self.last_load_mtime is None:
      return True
    try:
      return self.filename.stat().st_mode > self.last_load_mtime
    except FileNotFoundError:
      return True




import os
import subprocess as sp
import typing as t


class Branch(t.NamedTuple):
  name: str
  current: bool


class FileStatus(t.NamedTuple):
  mode: str
  filename: str


class RefWithSha(t.NamedTuple):
  ref: str
  sha: str


class Remote(t.NamedTuple):
  name: str
  fetch: str
  push: str


class Git:
  """
  Utility class to interface with the Git commandline.
  """

  def __init__(self, path: str = None):
    self.path = path or os.getcwd()

  def check_call(self, command: t.List[str], stdout: t.Optional[int] = None) -> None:
    sp.check_call(command, cwd=self.path, stdout=stdout)

  def check_output(self, command: t.List[str], stderr: t.Optional[int] = None) -> bytes:
    return sp.check_output(command, cwd=self.path, stderr=stderr)

  def init(self) -> None:
    self.check_call(['git', 'init', '.'])

  def clone(
      self,
      clone_url: str,
      branch: str = None,
      depth: int = None,
      recursive: bool = False,
      username: str = None,
      password: str = None,
      quiet: bool = False,
  ) -> None:
    """
    Clone a Git repository to the *to_directory* from *clone_url*. If a relative path is
    specified in *to_directory*, it will be treated relative to the #Git.cwd path.
    """

    if password or username:
      if not clone_url.startswith('https://'):
        raise ValueError('cannot specify username/password for non-HTTPS clone URL.')
      schema, remainder = clone_url.partition('://')[::2]
      auth = ':'.join(t.cast(t.List[str], filter(bool, [username, password])))
      clone_url = schema + '://' + auth + '@' + remainder

    command = ['git', 'clone', clone_url, self.path]
    if branch:
      command += ['-b', branch]
    if depth:
      command += ['--depth', str(depth)]
    if recursive:
      command += ['--recursive']
    if quiet:
      command += ['-q']

    # NOTE (NiklasRosenstein): We don't use #Git.check_call() as that would try to
    # change directory to the clone target directory, which might not yet exist.
    sp.check_call(command)

  def add(self, files: t.List[str]) -> None:
    """
    Add files to the index.
    """

    assert isinstance(files, list), f'expected list, got {type(files).__name__}'
    command = ['git', 'add', '--'] + files
    self.check_call(command)

  def get_branches(self) -> t.List[Branch]:
    """
    Get the branches of the repository. Returns a list of #Branch objects.
    """

    command = ['git', 'branch']
    results = []
    for line in self.check_output(command).decode().splitlines():
      current = False
      if line.startswith('*'):
        line = line[1:]
        current = True
      results.append(Branch(line.strip(), current))

    return results

  def get_branch_names(self) -> t.List[str]:
    """
    Get the branch names.
    """

    return [x.name for x in self.get_branches()]

  def get_current_branch_name(self) -> str:
    """
    Return the name of the current branch.
    """

    for branch in self.get_branches():
      if branch.current:
        return branch.name

    raise RuntimeError('no curent branch ?')

  def get_remote_refs(self, remote: str) -> t.List[RefWithSha]:
    result = []
    command = ['git', 'ls-remote', '--heads', 'origin']
    for line in self.check_output(command).decode().splitlines():
      sha, ref = line.split()
      result.append(RefWithSha(ref, sha))
    return result

  def get_remote_branch_names(self, remote: str) -> t.List[str]:
    refs = self.get_remote_refs(remote)
    return [x.ref[11:] for x in refs if x.ref.startswith('refs/heads/')]

  def rename_branch(self, current: str, new: str) -> None:
    self.check_call(['git', 'branch', '-m', current, new])

  def push(self, remote: str, *refs, force: bool = False) -> None:
    """
    Push the specified *refs* to the Git remote.
    """

    command = ['git', 'push', remote] + list(refs)
    if force:
      command.insert(2, '-f')
    self.check_call(command)

  def pull(self, remote: str = None, branch: str = None, quiet: bool = False):
    """
    Pull from the specified Git remote.
    """

    command = ['git', 'pull']
    if remote and branch:
      command += [remote, branch]
    elif remote or branch:
      raise ValueError('remote and branch arguments can only be specified together')
    if quiet:
      command += ['-q']

    self.check_call(command)

  def fetch(
    self,
    remote: str = None,
    all: bool = False,
    tags: bool = False,
    prune: bool = False,
    prune_tags: bool = False,
    argv: t.Optional[t.List[str]] = None
  ) -> None:
    """
    Fetch a remote repository (or multiple).
    """

    command = ['git', 'fetch']
    if remote:
      command += [remote]
    if all:
      command += ['--all']
    if tags:
      command += ['--tags']
    if prune:
      command += ['--prune']
    if prune_tags:
      command += ['--prune-tags']
    command += argv or []

    self.check_call(command)

  def remotes(self) -> t.List[Remote]:
    """
    List up all the remotes of the repository.
    """

    remotes: t.Dict[str, t.Dict[str, str]] = {}
    for line in self.check_output(['git', 'remote', '-v']).decode().splitlines():
      remote, url, kind = line.split()
      remotes.setdefault(remote, {})[kind] = url

    return [Remote(remote, urls['(fetch)'], urls['(push)']) for remote, urls in remotes.items()]

  def add_remote(self, remote: str, url: str, argv: t.Optional[t.List[str]] = None) -> None:
    """
    Add a remote with the specified name.
    """

    command = ['git', 'remote', 'add', remote, url] + (argv or [])
    self.check_call(command)

  def porcelain(self) -> t.Iterable[FileStatus]:
    """
    Returns the file status for the working tree.
    """

    for line in self.check_output(['git', 'status', '--porcelain']).decode().splitlines():
      mode, filename = line.strip().partition(' ')[::2]
      yield FileStatus(mode, filename)

  def commit(self, message: str, allow_empty: bool = False) -> None:
    """
    Commit staged files to the repository.
    """

    command = ['git', 'commit', '-m', message]
    if allow_empty:
      command.append('--allow-empty')
    self.check_call(command)

  def tag(self, tag_name: str, force: bool = False) -> None:
    """
    Create a tag.
    """

    command = ['git', 'tag', tag_name] + (['-f'] if force else [])
    self.check_call(command)

  def rev_parse(self, rev: str) -> t.Optional[str]:
    """
    Parse a Git ref into a shasum.
    """

    command = ['git', 'rev-parse', rev]
    try:
      return self.check_output(command, stderr=sp.STDOUT).decode().strip()
    except sp.CalledProcessError:
      return None

  def rev_list(self, rev: str, path: str = None) -> t.List[str]:
    """
    Return a list of all Git revisions, optionally in the specified path.
    """

    command = ['git', 'rev-list', rev]
    if path:
      command += ['--', path]
    try:
      revlist = self.check_output(command, stderr=sp.STDOUT).decode().strip().split('\n')
    except sp.CalledProcessError:
      return []
    if revlist == ['']:
      revlist = []
    return revlist

  def has_diff(self) -> bool:
    """
    Returns #True if the repository has changed files.
    """

    try:
      self.check_call(['git', 'diff', '--exit-code'], stdout=sp.PIPE)
      return False
    except sp.CalledProcessError as exc:
      if exc.returncode == 1:
        return True
      raise

  def create_branch(self, name: str, orphan: bool = False, reset: bool = False, ref: t.Optional[str] = None) -> None:
    """
    Creates a branch.
    """

    command = ['git', 'checkout']
    if orphan:
      if ref:
        raise ValueError('cannot checkout orphan branch with ref')
      command += ['--orphan', name]
    else:
      command += ['-B' if reset else '-b', name]
      if ref:
        command += [ref]

    self.check_call(command)

  def checkout(self, ref: str = None, files: t.List[str] = None, quiet: bool = False):
    """
    Check out the specified ref or files.
    """

    command = ['git', 'checkout']
    if ref:
      command += [ref]
    if quiet:
      command += ['-q']
    if files:
      command += ['--'] + files
    self.check_call(command)

  def reset(self, ref: str = None, files: t.List[str] = None, hard: bool = False, quiet: bool = False):
    """
    Reset to the specified ref or reset the files.
    """

    command = ['git', 'reset']
    if ref:
      command += [ref]
    if quiet:
      command += ['-q']
    if files:
      command += ['--'] + files
    self.check_call(command)

  def get_commit_message(self, rev: str) -> str:
    """
    Returns the commit message of the specified *rev*.
    """

    return self.check_output(['git', 'log', '-1', rev, '--pretty=%B']).decode()

  def get_diff(self, files: t.List[str] = None, cached: bool = False):
    command = ['git', '--no-pager', 'diff', '--color=never']
    if cached:
      command += ['--cached']
    if files is not None:
      command += ['--'] + files
    return self.check_output(command).decode()

  def describe(self,
    all: bool = False,
    tags: bool = False,
    contains: bool = False,
    commitish: t.Optional[str] = None,
  ) -> t.Optional[str]:

    command = ['git', 'describe']
    if all:
      command.append('--all')
    if tags:
      command.append('--tags')
    if contains:
      command.append('--contains')
    if commitish:
      command.append(commitish)

    try:
      return self.check_output(command, stderr=sp.DEVNULL).decode().strip()
    except FileNotFoundError:
      raise
    except sp.CalledProcessError:
      return None




"""
Provides generalized APIs to orchestrate background tasks.
"""

from ._api import Executor, Runnable, TaskStatus, TaskCallback, Task
from ._default import DefaultExecutor

__all__ = [
  'Executor',
  'Runnable',
  'TaskStatus',
  'TaskCallback',
  'Task',
  'DefaultExecutor',
]




"""
Utilities to mark values safe or unsafe, for example for the purpose of logging or to expose to users.
Note that anything accepting #Arg instances must explicitly implement support.
"""

import typing as t

from nr.util.generic import T


class Arg(t.Generic[T]):
  """ Base class for arguments. Cannot be directly constructed. """

  def __init__(self, value: T) -> None:
    if type(self) not in (Safe, Unsafe):
      raise RuntimeError(f'cannot create object of type {type(self).__name__}')
    self._value = value

  def __repr__(self) -> str:
    return f'{type(self).__name__}({self._value!r})'

  @property
  def value(self) -> T:
    return self._value

  def is_safe(self) -> t.TypeGuard['Safe']:
    return isinstance(self, Safe)

  def is_unsafe(self) -> t.TypeGuard['Unsafe']:
    return isinstance(self, Unsafe)


class Safe(Arg[T]):
  pass


class Unsafe(Arg[T]):
  pass




import abc
import enum
import logging
import types
import typing as t

from nr.util.generic import T

ExcInfoType = t.Tuple[t.Type[BaseException], BaseException, t.Optional[types.TracebackType]]
TaskCallback = t.Callable[['Task'], None]
TaskCallbackCondition = t.Callable[['Task'], bool]


class Runnable(abc.ABC, t.Generic[T]):
  """
  Abstract representation of something that can be run, for example in a task. Depending on the
  use case, runnables may need to be serializable.
  """

  @abc.abstractmethod
  def run(self, task: 'Task') -> T: ...


class TaskStatus(enum.Enum):
  """
  Represents the statuses that a task can be in.
  """

  #: The task was created but is not queued for execution.
  PENDING = 0

  #: The task is queued for execution.
  QUEUED = 1

  #: The task is currently running.
  RUNNING = 2

  #: The task has succeeded.
  SUCCEEDED = 3

  #: The task has failed (#Task.error will be set with the Python exception).
  FAILED = 4

  #: The task started but was cancelled while it was running. A task will only receive this
  #: status after it actually finished (i.e., #TaskStatus.cancelled() may return `True` while
  #: the #Task.status is still #RUNNING). A cancelled task may still have a #TaskStatus.result.
  CANCELLED = 5

  #: The task was queued but then ignored because the queue it was connected to was shut down.
  IGNORED = 6

  @property
  def idle(self) -> bool:
    """
    True if the status is either #PENDING or #QUEUED.
    """

    return self in (TaskStatus.PENDING, TaskStatus.QUEUED)

  @property
  def running(self) -> bool:
    """
    True if the status is #RUNNING.
    """

    return self == TaskStatus.RUNNING

  @property
  def done(self) -> bool:
    """
    True if the status is either #SUCCEEDED, #FAILED or #IGNORED.
    """

    return self in (TaskStatus.SUCCEEDED, TaskStatus.FAILED, TaskStatus.CANCELLED)

  @property
  def ignored(self) -> bool:
    """
    True if the status is #IGNORED.
    """

    return self == TaskStatus.IGNORED

  @property
  def completed(self) -> bool:
    """
    Returns `True` if the status represents a "completed" stated, i.e. the task will not change
    going forward (except maybe for it's #Task.error_consumed property). The set of completed statuses
    is #SUCCEEDED, #FAILED, #CANCELLED and #IGNORED.
    """

    return self in (TaskStatus.SUCCEEDED, TaskStatus.FAILED, TaskStatus.CANCELLED, TaskStatus.IGNORED)


class TaskCallbacks(abc.ABC):

  @abc.abstractmethod
  def add(
    self,
    condition: TaskCallbackCondition,
    callback: TaskCallback,
    once: bool = True,
    group: t.Optional[str] = None,
  ) -> None:
    """
    Add a callback to the task. If the *group* is specified, it can be used to remove the callbacks
    from the task wtih #remove_group(). The #condition determines when the *callback* will be invoked.
    If it returns `True` at the time when #add() is used, the *callback* is invoked immediately and
    will not be added to the collection.

    # Arguments
    condition: The condition on which to invoke them *callback*.
    callback: The callback to invoke when the task status is updated and the *condition* matches.
    once: If set to `True`, the callback will be invoked only once and removed from the task after.
    group: A string that identifies the group that the callback belongs to.
    """

  @abc.abstractmethod
  def remove(self, *, group: str) -> None:
    """
    Remove all callbacks from the collection that belong to the specified *group*.
    """

  def on(
    self,
    state: t.Union[TaskStatus, t.Sequence[TaskStatus], t.Literal['start'], t.Literal['end']],
    callback: TaskCallback,
    once: bool = True,
    group: t.Optional[str] = None,
  ) -> None:
    """
    Adds *callback* to the collection, which will be invoked if the task status is or changes to one
    that is specified via *state*. If *state* is `'end'`, it will match #TaskStatus.SUCCEEDED,
    #TaskStatus.FAILED and #TaskStatus.IGNORED. If the *state* is `'start'`, it behaves the same as
    # `'end'` but also match #TaskStatus.RUNNING.
    """

    statuses: t.Sequence[TaskStatus]
    if isinstance(state, str):
      end_statuses = (TaskStatus.SUCCEEDED, TaskStatus.FAILED, TaskStatus.IGNORED)
      if state == 'start':
        statuses = (TaskStatus.RUNNING,) + end_statuses
      elif state == 'end':
        statuses = end_statuses
      else:
        raise ValueError(f'invalid state: {state!r}')
    elif isinstance(state, TaskStatus):
      statuses = (state,)
    else:
      statuses = state

    def condition(t: 'Task') -> bool:
      return t.status in statuses

    self.add(condition, callback, once, group)


class Executor(abc.ABC):
  """
  Interface for task managers that can dispatch tasks for execution.
  """

  @abc.abstractmethod
  def get_worker_count(self) -> int: ...

  @abc.abstractmethod
  def get_idle_worker_count(self) -> int: ...

  @abc.abstractmethod
  def execute(
    self,
    runnable: Runnable,
    name: t.Optional[str] = None,
    at: t.Optional[float] = None,
  ) -> 'Task':
    """
    Execute the given *runnable* object in the task manager.

    # Arguments
    runnable: The runnable to invoke. Depending on the implementation of the task manager, this object
      may need to be serializable with whatever serializer the implementation is using (in case of a
      distributed task manager).
    name: An optional name for the runnable. If not specified, the #repr() of the object will be used.
    at: The timestamp at which the runnable is supposed to be executed. If the timestamp is smaller than
      the current time or if the parameter is not set, it will be executed immediately.

    # Returns
    The #Task object for this runnable.
    """

  @abc.abstractmethod
  def shutdown(self, cancel_running_taks: bool = True, block: bool = True) -> None:
    """
    Shut down the task manager, preventing new tasks from being exexuted. if *cancel_running_tasks*
    is enabled, all currently running tasks will be cancelled. By default, the method blocks until
    all running tasks have exited.

    This method may raise a #RuntimeError if it was called before.
    """

  @abc.abstractmethod
  def join(self) -> None:
    """
    Block until all pending and currently running tasks have been processed.
    """

  def idlejoin(self) -> None:
    """
    Like #join(), but garuantee a call to #shutdown() in the end.
    """

    try:
      self.join()
    finally:
      self.shutdown()


class Task(abc.ABC, t.Generic[T]):
  """
  Abstract representation of a task.
  """

  Status: t.ClassVar = TaskStatus
  Callback: t.ClassVar = TaskCallback
  Runnable: t.ClassVar = Runnable
  Executor: t.ClassVar = Executor

  @abc.abstractproperty
  def id(self) -> str: ...

  @abc.abstractproperty
  def worker_id(self) -> t.Optional[str]: ...

  @abc.abstractproperty
  def name(self) -> str: ...

  @abc.abstractproperty
  def logger(self) -> logging.Logger: ...

  @abc.abstractproperty
  def callbacks(self) -> TaskCallbacks: ...

  @abc.abstractproperty
  def status(self) -> TaskStatus: ...

  @abc.abstractproperty
  def error(self) -> t.Optional[ExcInfoType]: ...

  @abc.abstractproperty
  def error_consumed(self) -> bool: ...

  @abc.abstractproperty
  def result(self) -> t.Optional[T]:
    """
    Return the result value of the task. Raises a `RuntimeError` while the task is still running.
    Re-raises the #error if it is set. Returns `None` if the task has been ignored.
    """

  @abc.abstractmethod
  def consume_error(self, origin: t.Optional[str] = None) -> None: ...

  @abc.abstractmethod
  def cancel(self) -> None: ...

  @abc.abstractmethod
  def cancelled(self) -> bool:
    """
    Returns `True` if the task has been cancelled.
    """

  @abc.abstractmethod
  def sleep(self, duration: float) -> bool:
    """
    Sleep for *duration* seconds or until the task is cancelled.

    Returns `True` if the sleep completed, `False` is the timeout triggered which means
    that the task has been cancelled.
    """

  @abc.abstractmethod
  def join(self, timeout: t.Optional[float] = None) -> bool:
    """
    Block until the task completes or until the *timeout* is exceeded. A task is complete when its
    status is one of #TaskStatus.SUCCEEDED, #TaskStatus.FAILED or #TaskStatus.IGNORED. Returns `True`
    if the join is complete, `False` if the timeout was exceeded.
    """




"""
Provides the #DefaultTaskManager which operates locally to run tasks in a pool of threads.
"""

import dataclasses
import logging
import queue
import sys
import threading
import time
import typing as t
import uuid
import weakref

from . import _api as api
from nr.util.atomic import AtomicCounter
from nr.util.generic import T

logger = logging.getLogger(__name__)


class TaskCallbacks(api.TaskCallbacks):

  class _Entry(t.NamedTuple):
    condition: api.TaskCallbackCondition
    callback: api.TaskCallback
    once: bool
    group: t.Optional[str]

  def __init__(self, task: 'Task') -> None:
    self._task = weakref.ref(task)
    self._lock = task._lock
    self._callbacks: t.Dict[str, TaskCallbacks._Entry] = {}

  def __repr__(self) -> str:
    task = self._task()
    assert task is not None
    return f'<_TaskCallbacks task={task.name!r} size={len(self._callbacks)}>'

  def add(
    self,
    condition: api.TaskCallbackCondition,
    callback: api.TaskCallback,
    once: bool = True,
    group: t.Optional[str] = None,
  ) -> None:
    """
    Add a callback to the collection. If the *group* is specified, it can be used to remove
    the callbacks from the task wtih #remove_group(). The #condition determines when the *callback*
    will be invoked. If it returns `True` at the time when #add() is used, the *callback* is invoked
    immediately and will not be added to the collection.

    # Arguments
    condition: The condition on which to invoke them *callback*.
    callback: The callback to invoke when the task status is updated and the *condition* matches.
    once: If set to `True`, the callback will be invoked only once and removed from the task after.
    group: A string that identifies the group that the callback belongs to.
    """

    assert callable(callback)
    task = self._task()
    assert task is not None

    with self._lock:
      run_now = condition(task)
      if not run_now:
        self._callbacks[str(uuid.uuid4())] = TaskCallbacks._Entry(condition, callback, once, group)

    if run_now:
      callback(task)

  def remove(self, *, group: str) -> None:
    """
    Remove all callbacks from the collection that belong to the specified *group*.
    """

    with self._lock:
      self._callbacks = {k: e for k, e in self._callbacks.items() if e.group != group}

  def _invoke(self) -> None:
    """
    Internal. Invokes all callbacks based on their condition.
    """

    task = self._task()
    assert task is not None

    with self._lock:
      callbacks = self._callbacks.copy()

    remove: t.Set[str] = set()
    for key, entry in callbacks.items():
      try:
        do_run = entry.condition(task)
      except:
        task.logger.exception(f'Unhandled exception in callback condition of task "%s": %s', task.name, entry.condition)
        continue
      if not do_run:
        continue
      if entry.once:
        remove.add(key)
      try:
        entry.callback(task)
      except:
        task.logger.exception(f'Unhandled exception in callback of task "%s": %s', task.name, entry.callback)

    with self._lock:
      self._callbacks = {k: e for k, e in self._callbacks.items() if k not in remove}


class Task(api.Task[T]):

  def __init__(self, runnable: api.Runnable[T], name: str) -> None:
    self._lock = threading.RLock()
    self._cond = threading.Condition(self._lock)
    self._runnable = runnable
    self._id = type(self).__module__ + '.' + type(self).__name__ + '.' + str(uuid.uuid4())
    self._worker_id: t.Optional[str] = None
    self._name = name
    self._logger = self._create_logger()
    self._callbacks = TaskCallbacks(self)
    self._cancelled = threading.Event()
    self._status = api.TaskStatus.PENDING
    self._error: t.Optional[api.ExcInfoType] = None
    self._error_consumed: bool = False
    self._result: t.Optional[T] = None

  def _create_logger(self) -> logging.Logger:
    fqn = type(self).__module__ + '.' + type(self).__name__
    return logging.getLogger(f'{fqn}[{self.name}]')

  def _update(self, status: api.TaskStatus, result: t.Optional[T] = None, error: t.Optional[api.ExcInfoType] = None) -> None:
    """
    Update the status of the task. This should be used only by the executor engine where the task
    is queued. A status change will immediately invoke the registered #callbacks. Some status
    transitions are not allowed (ex. from #api.TaskStatus.SUCCEEDED to #api.TaskStatus.RUNNING). In this
    case, a #RuntimeError will be raised. Similarly, if the *status* is #api.TaskStatus.FAILED but
    no *error* is given, a #RuntimeError will be raised as well.
    """

    if status == api.TaskStatus.FAILED and error is None:
      raise RuntimeError(f'missing error information for setting task status to FAILED')

    if result is not None and status != api.TaskStatus.SUCCEEDED:
      raise RuntimeError(f'cannot set result with status {status.name}')

    with self._cond:
      if (self._status.completed and self._status != status) or (self._status == api.TaskStatus.RUNNING and status.idle):
        raise RuntimeError(f'changing the task status from {self._status.name} to {status.name} is not allowed')
      invoke_callbacks = status != self._status
      self._status = status
      self._error = error
      self._result = result
      self._cond.notify_all()

    if invoke_callbacks:
      self.callbacks._invoke()

  @property
  def runnable(self) -> api.Runnable: return self._runnable

  @property
  def id(self) -> str: return self._id

  @property
  def worker_id(self) -> t.Optional[str]: return self._worker_id

  @property
  def name(self) -> str: return self._name

  @property
  def logger(self) -> logging.Logger: return self._logger

  @property
  def callbacks(self) -> TaskCallbacks: return self._callbacks

  @property
  def status(self) -> api.TaskStatus:
    """
    Returns the current status of the task.
    """

    with self._lock:
      return self._status

  @property
  def error(self) -> t.Optional[api.ExcInfoType]:
    """
    Returns the exception that occurred while executing the task. This is only set if
    the #status is #api.TaskStatus.FAILED.
    """

    with self._lock:
      return self._error

  @property
  def error_consumed(self) -> bool:
    """
    Returns #True if the error in the task is marked as consumed.
    """

    with self._lock:
      return self._error_consumed

  @property
  def result(self) -> t.Optional[T]:
    with self._lock:
      if self._status == api.TaskStatus.IGNORED:
        return None
      elif self._status == api.TaskStatus.FAILED:
        assert self._error is not None, 'Task status is FAILED but no error is set'
        raise self._error[1]
      elif self._status in (api.TaskStatus.SUCCEEDED, api.TaskStatus.CANCELLED):
        return self._result
      else:
        raise RuntimeError(f'Task has status {self._status.name}')

  def consume_error(self, origin: t.Optional[str] = None) -> None:
    """
    Mark the error in the task as consumed. This can be called from a callback, but can only be
    called if the #status is #api.TaskStatus.FAILED.
    """

    with self._lock:
      if self._status != api.TaskStatus.FAILED:
        raise RuntimeError(f'task status must be FAILED to call Task.consume_error() but is {self._status.name}')
      self._error_consumed = True

  def cancel(self) -> None:
    """
    Set the cancelled flag on the task. After calling this method, #cancelled() will return `True`
    and any thread that is currently using #sleep() will be immediately woken up (rather than
    waiting for the timeout to kick in).
    """

    self._cancelled.set()

  def cancelled(self) -> bool:
    """
    Returns `True` if the task has been cancelled (with the #cancel() methdo). The task subclass
    should use this to check if execution should continue or not.
    """

    if self._cancelled is None:
      raise RuntimeError('Task is not connected to a worker')
    return self._cancelled.is_set()

  def sleep(self, duration: float) -> bool:
    """
    Sleep for *sec* seconds, or until the task is cancelled with the #cancel() method. This should
    be used instead of #time.sleep() inside the task's #run() implementation to ensure quick
    task termination.

    Returns `True` if the task has been cancelled (saving a subsequent call to #cancelled()).
    """

    if self._cancelled is None:
      raise RuntimeError('Task is not connected to a worker')
    return not self._cancelled.wait(duration)

  def join(self, timeout: t.Optional[float] = None) -> bool:
    with self._cond:
      return self._cond.wait_for(lambda: self._status.completed, timeout)


class TaskPriorityQueue:

  def __init__(self) -> None:
    self._q: 'queue.PriorityQueue[t.Tuple[float, t.Optional[Task]]]' = queue.PriorityQueue()
    self._max_time: t.Optional[float] = None
    self._cond = threading.Condition(threading.Lock())
    self._total = AtomicCounter()
    self._current = AtomicCounter()

  def total(self) -> int: return self._total.get()

  def pending(self) -> int: return max(0, self._total.get() - self._current.get())

  def current(self) -> int: return self._current.get()

  def max_time(self) -> t.Optional[float]:
    with self._cond:
      return self._max_time

  def task_done(self) -> None:
    self._q.task_done()
    self._current.dec()
    self._total.dec()

  def put(self, at: float, task: Task) -> None:
    assert isinstance(at, float)
    assert isinstance(task, Task)
    with self._cond:
      self._q.put((at, task))
      self._total.inc()
      self._max_time = at if self._max_time is None else max (at, self._max_time)
      self._cond.notify()

  def put_stop(self, at: float) -> None:
    with self._cond:
      self._q.put((at, None))
      self._cond.notify_all()

  def get(self) -> t.Optional[Task]:
    while True:
      at, task = self._q.get()
      if task is None:
        self._q.task_done()
        return None

      t = time.time()
      if at <= t or task.cancelled():
        self._current.inc()
        return task

      with self._cond:
        self._q.task_done()
        self._q.put((at, task))
        self._cond.wait(timeout=at - t)

  def join_total(self) -> None:
    self._total.join()

  def join_current(self) -> None:
    self._current.join()


@dataclasses.dataclass
class Worker:
  """
  The worker thread accepts tasks from a queue and executes them. Receiving #None from the queue
  indicates to the worker that its should stop.
  """

  #: The name of the worker is used as the thread name.
  name: str

  #: The queue to retrieve new tasks from.
  queue: TaskPriorityQueue

  def __post_init__(self) -> None:
    self._thread = threading.Thread(target=self._run, name=self.name, daemon=True)
    self._lock = threading.Lock()
    self._current_task: t.Optional[Task] = None

  def __repr__(self) -> str:
    return f'Worker(name={self.name!r})'

  def _run(self) -> None:
    logger.info('Start worker %s', self.name)
    try:
      self._mainloop()
    except:
      logging.exception('Unhandled exception in worker "%s"', self.name)
    logger.info('Stopped worker %s', self.name)

  def _mainloop(self) -> None:
    while True:
      task = self.queue.get()
      if task is None:
        break
      if task.cancelled():
        task._update(api.TaskStatus.IGNORED)
        self.queue.task_done()
        continue
      with self._lock:
        self._current_task = task
      try:
        self._run_task(task)
      finally:
        with self._lock:
          self._current_task = None
        self.queue.task_done()

  def _run_task(self, task: Task) -> None:
    logger.info('Running task "%s"', task.name)

    try:
      task._worker_id = self.name
      task._update(api.TaskStatus.RUNNING)
      result = task.runnable.run(task)
    except:
      task._update(api.TaskStatus.FAILED, t.cast(api.ExcInfoType, sys.exc_info()))
      if not task._error_consumed:
        logger.exception('Unhandled exception in task "%s"', task.name)
    else:
      task._update(api.TaskStatus.CANCELLED if task.cancelled() else api.TaskStatus.SUCCEEDED, result)
    finally:
      logger.info('Finished task "%s"', task.name)

  def get_current_task(self) -> t.Optional[Task]:
    with self._lock:
      return self._current_task

  def start(self) -> None:
    self._thread.start()


@dataclasses.dataclass  # type: ignore
class DefaultExecutor(api.Executor):

  _CALLBACK_GROUP = '_TaskManager'

  #: The name of the task manager. This is used as the prefix for created #Worker names.
  name: str

  #: The maximum number of workers to spawn. Defaults to 8,
  max_workers: int = 8

  def __post_init__(self) -> None:
    #: A list of the workers assigned to the pool that is bounded by #max_size.
    self._pool_workers: t.List[Worker] = []

    #: The queue that is used to send tasks to the workers.
    self._queue = TaskPriorityQueue()

    #: A counter for all tasks submitted to the task manager that have not yet completed.
    self._all_tasks = AtomicCounter()

    #: A counter for all currently active tasks.
    self._active_tasks = AtomicCounter()

    self._lock = threading.Lock()
    self._shutdown = False

  def execute(
    self,
    runnable: api.Runnable[T],
    name: t.Optional[str] = None,
    at: t.Optional[float] = None,
  ) -> Task[T]:
    """
    Queue a task for execution in the worker pool. If there is a free slot in the pool and all
    existing workers are busy, a new worker will be spawned immediately.
    """

    assert isinstance(runnable, api.Runnable), f'expected instance of Runnable, got {type(runnable).__name__} instead'
    if self._shutdown:
      raise RuntimeError('task manager is shut down')

    task = Task(runnable, name or repr(runnable))
    task._update(api.TaskStatus.QUEUED)
    self._queue.put(at or time.time(), task)
    size = self._all_tasks.get()
    if size >= len(self._pool_workers) and size < self.max_workers:
      self._spawn_pool_worker()
    return task

  def _spawn_pool_worker(self) -> None:
    worker = Worker(f'{self.name}-Worker-{len(self._pool_workers)}', self._queue)
    worker.start()
    self._pool_workers.append(worker)

  def get_worker_count(self) -> int:
    return len(self._pool_workers)

  def get_idle_worker_count(self) -> int:
    return len(self._pool_workers) - self._queue.current()

  def shutdown(self, cancel_running_tasks: bool = True, block: bool = True) -> None:
    if self._shutdown:
      raise RuntimeError('shut down already initiated or completed')
    self._shutdown = True

    logger.info('Sending shutdown signal to workers')

    for worker in self._pool_workers:
      if cancel_running_tasks:
        task = worker.get_current_task()
        if task is not None:
          task.cancel()
      # Use the current time to jump in front of pending tasks.
      self._queue.put_stop(0)

    if block:
      self._queue.join_current()

  def join(self) -> None:
    self._queue.join_total()
